
/*
 * libcookie - a HTTP cookie library for ANSI C
 * Copyright (C) 1998-1999 Alex Demko <ademko@mbnet.mb.ca>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * cookie.c - the libcookie cookie library
 * $Id: cookie.c,v 1.11 1999/03/06 17:19:46 ademko Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <cookie.h>

/* a cookie node on the internal linked list */
typedef struct cookienode {
   char *name, *value;
   struct cookienode *next;
} cookienode, *pcookienode;

/* the head pointer to the linked list of cookies */
static cookienode *cookies;

/* various 'set' parameters */
static char
   cookieexpires[COOKIE_STRING_LENGTH],
   cookiedomain[COOKIE_STRING_LENGTH],
   cookiepath[COOKIE_STRING_LENGTH];

/* secure connection required? */
static int cookiesecure;

/* the stream which all routines write out too */
static FILE *cookiestdout = stdout;

/* internal routine used to set the fields */
static char * s_set(char *stored, const char *new_s) {
   char *old;

   if (!new_s)
      return stored;
   
   old = stored;
   strncpy(stored,new_s,COOKIE_STRING_LENGTH);
   return old;
}

/* can a character be sent without escaping it? */
static int ok_char(char c) {
   return isalnum(c) || (c=='_') || (c=='-');
}

/* convert a hex ascii char to its int value */
static int hex_to_int(char hexchar) {
   if ((hexchar>='0') && (hexchar<='9'))
      return hexchar - '0';
   if ((hexchar>='a') && (hexchar<='f'))
      return hexchar - 'a' + 0x0a;
   if ((hexchar>='A') && (hexchar<='F'))
      return hexchar - 'A' + 0x0a;
   return 0;
}

/* int to hex char map */
static const char hex_map[]="0123456789ABCDEF";

/* escapes a string */
static void escape_string(char *dest, const char *src) {
   const char *csrc;
   char *cdest;
   int max;

   /* make sure both strings were sent */
   if (!dest || !src)
      return;

   max=COOKIE_STRING_LENGTH-5;
   csrc=src;
   cdest=dest;

   /* interate through the whole string */
   while ((max>0) && *csrc) {
      if (ok_char(*csrc)) {
	 /* char does not need to be encoded */
         *cdest=*csrc;
	 cdest++;
	 max--;
      } else if (*csrc == ' ') {
	 /* special case encode - space */
         *cdest='+';
	 cdest++;
	 max--;
      } else {
         /* must encode this char */
	 *cdest='%';
	 cdest++;
	 *cdest=hex_map[(*csrc & 0xf0) >> 4];
	 cdest++;
	 *cdest=hex_map[(*csrc & 0x0f)];
	 cdest++;
	 max-=3;
      }
      csrc++;
   }
   *cdest=0;
}

/* unescape a string */
static void unescape_string(char *dest, const char *src) {
   const char *csrc;
   char *cdest, c;
   int max;

   max=COOKIE_STRING_LENGTH;
   csrc=src;
   cdest=dest;

   /* interate through the whole string */
   while (max && *csrc) {
      if (*csrc == '+') {
         /* decode space */
	 *cdest = ' ';
	 csrc++;
      } else if (*csrc == '%') {
         /* escape char - decode hex number */
         csrc++;
	 c=hex_to_int(*csrc) << 4;
	 if (*csrc) {
	    csrc++;
	    c|=hex_to_int(*csrc);
	    csrc++;
	 }
	 *cdest=c;
      } else {
         /* ok char - just copy */
	 *cdest = *csrc;
	 csrc++;
      }
      cdest++;
      max--;
   }
   *cdest=0; /* terminate the string */
}

/* adds a cookie to the iternal linked list */
static int addcookie(char *pair) {
   char *value, ename[COOKIE_STRING_LENGTH],
       evalue[COOKIE_STRING_LENGTH];
   cookienode *node;

   /* find the split point */
   value = strchr(pair,'=');
   if (value) { 
        *value=0;
        value++;
        unescape_string(ename,pair);
        unescape_string(evalue,value);
    } else unescape_string(ename,pair);

   if (!ename[0] || !evalue[0])
      return 0;

   /* add the new node */
   node = malloc(sizeof(cookienode));
   node->next=cookies;
   node->name=strdup(ename);
   node->value=strdup(evalue);
   cookies = node;

   return 1;
}

/* load the cookies from the HTTP_COOKIE enviromental variable
   into the internal linked list */
int cookieLoad(void) {
   char *cookieenv, *cur, buf[1+COOKIE_STRING_LENGTH], *cbuf;
   int count, max;

   cookieenv=getenv("HTTP_COOKIE");
   if (!cookieenv)
      return 0; /* env var not set - no cookies */

   /* clean the list - just in case */
   cookieFree();

   cur=cookieenv;
   max=COOKIE_STRING_LENGTH;
   count=0;
   cbuf=buf;
   /* loop through the whole string */
   while (max && *cur) {
      if (*cur == ';') {
        *cbuf=0;
        count += addcookie(buf);
        cbuf=buf;
      } else if (!((*cur==' ') && (buf==cbuf))) {
        *cbuf=*cur;
        cbuf++;
      }
      cur++;
      max--;
   }
   *cbuf=0;
   count += addcookie(buf);

   return count;
}

/* return the value of a particular cookie */
char * cookieGet(const char *cookiename) {
   cookienode *cur=cookies;

   /* search through the list */
   while (cur && strcmp(cookiename,cur->name))
      cur=cur->next;

   return (cur?cur->value:NULL);
}

/* free the linked list */
void cookieFree(void) {
   cookienode *t;

   /* interate through all the cookies while disposing of
      the dunamic strings */
   while (cookies) {
      t=cookies;
      cookies=cookies->next;
      free(t->name);
      free(t->value);
      free(t);
   }
}

/* open an interation list */
void * cookieListOpen(void) {
   return cookies;
}

/* extract the name from the current location of the list */
char * cookieListName(void *handle) {
   return (handle?((cookienode *)handle)->name:NULL);
}

/* extract the value from the current location of the list */
char * cookieListValue(void *handle) {
   return (handle?((cookienode *)handle)->value:NULL);
}

/* advance the list pointer */
void * cookieListNext(void *handle) {
   if (handle)
      return ((cookienode *)handle)->next;
      else return NULL;
}

/* set the value of a particular cookie */
void cookieSet(const char *name, const char *value) {
   char ename[COOKIE_STRING_LENGTH], evalue[COOKIE_STRING_LENGTH],
        expires[COOKIE_STRING_LENGTH], domain[COOKIE_STRING_LENGTH],
        path[COOKIE_STRING_LENGTH], buf[COOKIE_STRING_LENGTH];

   escape_string(ename,name);
   escape_string(evalue,value);

   /* check if we need to include the expires tag */
   if (cookieexpires[0]) {
      escape_string(buf,cookieexpires);
      sprintf(expires," expires=%s;",buf);
   } else expires[0]='\0';

   /* check if we need to include the domain tag */
   if (cookiedomain[0]) {
      escape_string(buf,cookiedomain);
      sprintf(domain," domain=%s;",buf);
   } else domain[0]='\0';

   /* check if we need to include the path tag */
   if (cookiepath[0]) {
      escape_string(buf,cookiepath);
      sprintf(path," path=%s;",buf);
   } else path[0]='\0';

   /* write the cookie set string out to our stream */
   fprintf(cookiestdout,"Set-Cookie: %s=%s;%s%s%s%s\n",
      ename,evalue,expires,domain,path,
      (cookiesecure?" secure":""));
}

/* set the cookie expirey data - must be in the form:
   DD-Mon-YYYY HH:MM:SS GMT */
char * cookieSetExpires(const char *newdate) {
   return s_set(cookieexpires, newdate);
}

/* set the domain of the cookies */
char * cookieSetDomain(const char *newdomain) {
   return s_set(cookiedomain, newdomain);
}

/* set the receiving path script path */
char * cookieSetPath(const char *newpath) {
   return s_set(cookiepath, newpath);
}

/* set if you want to require a secure link */
int cookieSetSecure(int needsecure) {
   int old = cookiesecure;

   cookiesecure=needsecure?1:0;
   return old;
}

/* check if you set the secure link requirement */
int cookieGetSecure(void) {
   return cookiesecure;
}

/* set a new outboud stream */
FILE * cookieSetOutFile(FILE *newfile) {
   FILE *old=cookiestdout;

   cookiestdout = newfile;

   return old;
}

